{=========================================================================
  RkMail Unit

  This unit implements the RkMailMessage component which uses the Simple
  Messaging API (MAPI) functions from MAPI32.DLL to send mail messages.
  This component supports multiple TO and CC recipients and multiple
  attached files.

  Developing Custom Delphi 3 Components
  Copyright  1995-1997 by Ray Konopka
=========================================================================}

{$I DCDC.INC}

unit RkMail;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, RkCommon;

type
  EMapiUserAbort = class( EAbort );                    // Silent exception
  EMapiError = class( Exception )
    ErrorCode : Integer;
  end;

  TRkMailMessage = class( TComponent )
  private
    FAboutInfo : TRkAboutInfo;
    FAttachments : TStrings;
    FToRecipients : TStrings;
    FCcRecipients : TStrings;
    FMessageMemo : TCustomMemo;
    FMessageText : TStrings;
    FPassword : string;
    FProfileName : string;
    FReview : Boolean;
    FSession : Longint;
    FSubject : string;
    FSubjectEdit : TCustomEdit;
    procedure SetAttachments( Value : TStrings );
    procedure SetToRecipients( Value : TStrings );
    procedure SetCcRecipients( Value : TStrings );
    procedure SetMailMessage( Value : TStrings );
    procedure SetMessageMemo( Value : TCustomMemo );
    procedure SetSubjectEdit( Value : TCustomEdit );
  protected
    procedure Notification( AComponent : TComponent;
                            Operation : TOperation ); override;
  public
    constructor Create( AOwner : TComponent ); override;
    destructor Destroy; override;

    procedure Logon;
    procedure Logoff;
    procedure Send;
  published
    property About : TRkAboutInfo
      read FAboutInfo
      write FAboutInfo
      stored False;

    property Attachments : TStrings
      read FAttachments
      write SetAttachments;

    property CcRecipients : TStrings
      read FCcRecipients
      write SetCcRecipients;

    property MessageMemo : TCustomMemo
      read FMessageMemo
      write SetMessageMemo;

    property MessageText : TStrings
      read FMessageText
      write SetMailMessage;

    property ProfileName : string
      read FProfileName
      write FProfileName;

    property Password : string
      read FPassword
      write FPassword;

    property Review : Boolean
      read FReview
      write FReview
      default True;

    property ToRecipients : TStrings
      read FToRecipients
      write SetToRecipients;

    property Subject : string
      read FSubject
      write FSubject;

    property SubjectEdit : TCustomEdit
      read FSubjectEdit
      write SetSubjectEdit;
  end;


implementation


{$IFDEF CHAPTER_11}
uses
  Mapi;

const
  MapiErrMsgs: array[mapi_User_Abort..mapi_E_Not_Supported] of string[24]=
    ( 'User Abort',
      'Failure',
      'Login Failure',
      'Disk Full',
      'Insufficient Memory',
      'Access Denied',
      '',
      'Too Many Sessions',
      'Too Many Files',
      'Too Many Recipients',
      'Attachment Not Found',
      'Attachment Open Failure',
      'Attachment Write Failure',
      'Unknown Recipient',
      'Bad RecipType',
      'No Messages',
      'Invalid Message',
      'Text Too Large',
      'Invalid Session',
      'Type Not Supported',
      'Ambiguous Recipient',
      'Message In Use',
      'Network Failure',
      'Invalid EditFields',
      'Invalid Recips',
      'Not Supported' );


function CreateMapiError( ErrCode : Integer ) : Exception;
begin
  if ErrCode = mapi_User_Abort then
  begin
    // If user aborts send operation, raise a silent exception
    Result := EMapiUserAbort.Create( 'MAPI: Process Aborted by User' );
  end
  else
  begin
    Result := EMapiError.CreateFmt( 'MAPI: %s. ErrorCode = %d',
                                    [ MapiErrMsgs[ ErrCode ], ErrCode ] );
    EMapiError( Result ).ErrorCode := ErrCode;
  end;
end;

{$ELSE}
uses
  Mapi, MapiMsgs;



function CreateMapiError( ErrCode : Integer ) : Exception;
begin
  if ErrCode = mapi_User_Abort then
  begin
    // If user aborts send operation, raise a silent exception
    Result := EMapiUserAbort.Create( LoadStr( SMapiUserAbort ) );
  end
  else
  begin
    Result := EMapiError.CreateFmt( 'MAPI: %s. ErrorCode = %d',
                                    [ LoadStr( SMapiBase + ErrCode ),
                                      ErrCode ] );
    EMapiError( Result ).ErrorCode := ErrCode;
  end;
end;
{$ENDIF}


{============================}
{== TRkMailMessage Methods ==}
{============================}


constructor TRkMailMessage.Create( AOwner : TComponent );
begin
  inherited Create( AOwner );
  FSession := 0;
  FAttachments := TStringList.Create;
  FToRecipients := TStringList.Create;
  FCcRecipients := TStringList.Create;
  FMessageText := TStringList.Create;
  FReview := True;
end;


destructor TRkMailMessage.Destroy;
begin
  FAttachments.Free;
  FToRecipients.Free;
  FCcRecipients.Free;
  FMessageText.Free;
  inherited Destroy;
end;


{=========================================================================
  TRkMailMessage.Notification

  This method is overridden to ensure that FMessageMemo and FSubjectEdit 
  are set to nil if the corresponding Memo component is deleted from the 
  form. This method should be used whenever a contains a reference to 
  another component.
=========================================================================}

procedure TRkMailMessage.Notification( AComponent : TComponent;
                                       Operation : TOperation );
begin
  inherited Notification( AComponent, Operation );

  if Operation = opRemove then
  begin
    if AComponent = FMessageMemo then
      FMessageMemo := nil
    else if AComponent = FSubjectEdit then
      FSubjectEdit := nil;
  end;
end;


procedure TRkMailMessage.SetAttachments( Value : TStrings );
begin
  FAttachments.Assign( Value );
end;


procedure TRkMailMessage.SetToRecipients( Value : TStrings );
begin
  FToRecipients.Assign( Value );
end;


procedure TRkMailMessage.SetCcRecipients( Value : TStrings );
begin
  FCcRecipients.Assign( Value );
end;


procedure TRkMailMessage.SetMailMessage( Value : TStrings );
begin
  FMessageText.Assign( Value );
end;


procedure TRkMailMessage.SetMessageMemo( Value : TCustomMemo );
begin
  FMessageMemo := Value;

  // Put the current component on the FFreeNotifies list of the specified
  // memo component by calling the memo's FreeNotification method.

  if FMessageMemo <> nil then
    FMessageMemo.FreeNotification( Self );
end;


procedure TRkMailMessage.SetSubjectEdit( Value : TCustomEdit );
begin
  FSubjectEdit := Value;

  // Put the current component on the FFreeNotifies list of the specified
  // edit component by calling the edit's FreeNotification method.

  if FSubjectEdit <> nil then
    FSubjectEdit.FreeNotification( Self );
end;


procedure TRkMailMessage.Logon;
var
  RetCode : Integer;
  Flags : Longint;
begin
  if FProfileName = '' then
    Flags := mapi_Logon_UI or mapi_Dialog or mapi_Use_Default
  else
    Flags := mapi_Logon_UI or mapi_Dialog;

  RetCode := MapiLogon( 0, PChar( FProfileName ), PChar( FPassword ),
                        Flags, 0, @FSession );
  if RetCode > 0 then
    raise CreateMapiError( RetCode );
end;


procedure TRkMailMessage.Logoff;
var
  RetCode : Integer;
begin
  if FSession = 0 then
    Exit;

  RetCode := MapiLogoff( FSession, 0, 0, 0 );
  if RetCode > 0 then
    raise CreateMapiError( RetCode );
  FSession := 0;
end;


{ Supporting Constants and Types used to cast generic pointers }
const
  MaxNumFiles = 65520 div SizeOf( TMapiFileDesc );
  MaxNumRecips = 65520 div SizeOf( TMapiRecipDesc );

type
  TFileArray = array[ 0..MaxNumFiles - 1 ] of TMapiFileDesc;
  TRecipArray = array[ 0..MaxNumRecips - 1 ] of TMapiRecipDesc;
  TRecipBufArray = array[ 0..MaxNumRecips - 1 ] of PMapiRecipDesc;


procedure TRkMailMessage.Send;
var
  AttachmentPositions : string;
  Files : Pointer;
  FilesMemSize : Word;
  FullPath : string;
  TempStz : array[ 0..255 ] of Char;
  RecipClass : Longint;
  Recips : Pointer;
  RecipsMemSize : Word;
  RecipBuffer : Pointer;
  RecipBufMemSize : Word;
  RecipName : array[ 0..255 ] of Char;
  Msg : TMapiMessage;
  RetCode : Longint;
  SubjectStz : array[ 0..255 ] of Char;
  I, J, MessageLen : Integer;
  SendFlags : Longint;
begin
  Screen.Cursor := crHourGlass;
  try
    if FSession = 0 then              // If not yet logged on, then log on
      Logon;

    FillChar( Msg, SizeOf( TMapiMessage ), 0 );    // Clear out Msg record

    if FMessageMemo <> nil then          // If component linked to memo...
      FMessageText.Assign( FMessageMemo.Lines );      // use memo contents

    // Attached files replace characters in the message body.  The
    // following code adds blank spaces at the end of the message so that
    // any attached { files will have a place to be inserted into the
    // message, without overwriting any portion of the message text.

    AttachmentPositions := '';
    for I := 1 to FAttachments.Count do
      AttachmentPositions := AttachmentPositions + ' ';
    if AttachmentPositions <> '' then
      FMessageText.Add( AttachmentPositions );

    // Populate the message body

    Msg.lpszNoteText := PChar( FMessageText.Text );
    MessageLen := Length( FMessageText.Text );

    if FSubjectEdit <> nil then
      StrPCopy( SubjectStz, FSubjectEdit.Text )
    else
      StrPCopy( SubjectStz, FSubject );
    Msg.lpszSubject := SubjectStz;

    // Add Recipients to Message

    // Specify the number of recipients
    Msg.nRecipCount := FToRecipients.Count + FCcRecipients.Count;
    RecipsMemSize := SizeOf( TMapiRecipDesc ) * Msg.nRecipCount;
    Recips := nil;
    RecipBufMemSize := SizeOf( PMapiRecipDesc ) * Msg.nRecipCount;
    RecipBuffer := nil;

    if Msg.nRecipCount > 0 then
    begin
      // Allocate memory for Recips array
      GetMem( Recips, RecipsMemSize );
      FillChar( Recips^, RecipsMemSize, 0 );

      // Allocate Memory for RecipBuffer
      GetMem( RecipBuffer, RecipBufMemSize );

      for I := 0 to Msg.nRecipCount - 1 do
      begin
        if I < FToRecipients.Count then
        begin
          StrPCopy( RecipName, FToRecipients[ I ] );
          RecipClass := mapi_TO;
        end
        else
        begin
          StrPCopy( RecipName, FCcRecipients[ I - FToRecipients.Count ] );
          RecipClass := mapi_CC;
        end;

        // Call MapiResolveName to get the Address for the
        // RecipName recipient

        RetCode := MapiResolveName( FSession, 0, RecipName,
                                    mapi_Logon_UI + mapi_Dialog, 0,
                                    TRecipBufArray( RecipBuffer^ )[ I ] );
        if RetCode > 0 then
        begin
          for J := 0 to I - 1 do
            MapiFreeBuffer( TRecipBufArray( RecipBuffer^ )[ J ] );
          raise CreateMapiError( RetCode );
        end;

        // Populate Recips array with data retrieved in RecipBuffer

        TRecipArray( Recips^ )[ I ].ulRecipClass := RecipClass;

        TRecipArray( Recips^ )[ I ].lpszName :=
          TRecipBufArray( RecipBuffer^ )[ I ]^.lpszName;

        TRecipArray( Recips^ )[ I ].lpszAddress :=
          TRecipBufArray( RecipBuffer^ )[ I ]^.lpszAddress;
      end;
      Msg.lpRecips := Recips;
    end;

    // Add Attachments to Message

    // Specify the number of attachments
    Msg.nFileCount := FAttachments.Count;
    Files := nil;
    FilesMemSize := SizeOf( TMapiFileDesc ) * FAttachments.Count;
    if Msg.nFileCount > 0 then
    begin
      GetMem( Files,  FilesMemSize );
      FillChar( Files^, FilesMemSize, 0 );

      for I := 0 to FAttachments.Count - 1 do
      begin
        FullPath := ExpandFileName( FAttachments[ I ] );
        TFileArray( Files^ )[ I ].lpszPathName :=
          StrNew( StrPCopy( TempStz, FullPath ) );

        // Place each attachment at one of the spaces added to the
        // end of the msg
        TFileArray( Files^ )[ I ].nPosition := MessageLen - 3 - I;
      end;
      Msg.lpFiles := Files;
    end;

    // Send the Message

    try
      if FReview or ( FToRecipients.Count = 0 ) then
        SendFlags := mapi_Logon_UI + mapi_Dialog   // Show the compose dlg
      else
        SendFlags := mapi_Logon_UI;

      // MapiSendMail will attempt to use the session established by
      // MapiLogon or if FSession is 0, it will use a shared session, and
      // if one is not available, a new session is started.

      RetCode := MapiSendMail( FSession, 0, Msg, SendFlags, 0 );
      if RetCode > 0 then
        raise CreateMapiError( RetCode );
    finally
      // Clean Up:  Lots of dynamic memory to free up

      if Msg.nRecipCount > 0 then
      begin
        for I := 0 to FToRecipients.Count + FCcRecipients.Count - 1 do
          MapiFreeBuffer( TRecipBufArray( RecipBuffer^ )[ I ] );
        FreeMem( Recips, RecipsMemSize );
        FreeMem( RecipBuffer, RecipBufMemSize );
      end;

      if Msg.nFileCount > 0 then
      begin
        for I := 0 to FAttachments.Count - 1 do
          StrDispose( TFileArray( Files^ )[ I ].lpszPathName );
        FreeMem( Files, FilesMemSize );
      end;
    end;
  finally
    Screen.Cursor := crDefault;
  end;
end;


end.
