{=========================================================================
  RkStatus Unit

  This unit provides several status components including one that
  monitors the system clock, and another that shows keyboard status.

  Developing Custom Delphi 3 Components
  Copyright  1995-1997 by Ray Konopka
=========================================================================}

{$I DCDC.INC}

unit RkStatus;

interface

uses
  Messages, Windows, Controls, Classes, Graphics, ExtCtrls, Menus,
  RkCommon, Buttons, SysUtils;

type
  TRkCustomStatusControl = class( TGraphicControl )
  private
    FAboutInfo : TRkAboutInfo;
    FAlignment : TAlignment;
    FBorderWidth : TBorderWidth;
    FCaption : string;
    FCaptionOffset : Integer;
    FFrameStyle : TFrameStyle;

    procedure SetAlignment( Value : TAlignment );
    procedure SetBorderWidth( Value : TBorderWidth );
    procedure SetCaption( Value : string );
    procedure SetCaptionOffset( Value : Integer );
    procedure SetFrameStyle( Value : TFrameStyle );
    procedure CMTextChanged( var Msg : TMessage ); message cm_TextChanged;
  protected
    function GetCaptionRect : TRect; virtual;
    procedure DrawCaption( CaptionStr : string );
    procedure DrawStatusBorder;
    procedure Paint; override;

    property Alignment : TAlignment
      read FAlignment
      write SetAlignment
      default taLeftJustify;

    property CaptionOffset : Integer
      read FCaptionOffset
      write SetCaptionOffset
      default 0;

    property Caption : string                        // Redefined Property
      read FCaption
      write SetCaption;
  public
    constructor Create( AOwner : TComponent ); override;
  published
    property About : TRkAboutInfo
      read FAboutInfo
      write FAboutInfo
      stored False;

    property FrameStyle : TFrameStyle
      read FFrameStyle
      write SetFrameStyle
      default fsRecessed;

    property BorderWidth : TBorderWidth
      read FBorderWidth
      write SetBorderWidth
      default 2;

    { Inherited Properties }
    property Align;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property Height default 20;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Visible;
    property Width default 100;

    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnStartDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end; {== TRkCustomStatusControl Class Declaration ==}


  TRkStatusControl = class( TRkCustomStatusControl )
  published
    property Alignment;
    property Caption;
    property CaptionOffset;
  end;


  TGlyphAlignment = ( gaLeft, gaRight );

  TRkCustomGlyphStatus = class( TRkCustomStatusControl )
  private
    FGlyph : TBitmap;
    FGlyphAlignment : TGlyphAlignment;
    FGlyphOffset : Integer;
    FNumGlyphs : TNumGlyphs;

    procedure SetNumGlyphs( Value : TNumGlyphs );
    procedure SetGlyph( Value : TBitmap );
    procedure SetGlyphAlignment( Value : TGlyphAlignment );
    procedure SetGlyphOffset( Value : Integer );
  protected
    procedure GlyphChanged( Sender : TObject );
    function GetCaptionRect : TRect; override;
    procedure Paint; override;

    property Glyph : TBitmap
      read FGlyph
      write SetGlyph;

    property GlyphAlignment : TGlyphAlignment
      read FGlyphAlignment
      write SetGlyphAlignment
      default gaLeft;

    property GlyphOffset : Integer
      read FGlyphOffset
      write SetGlyphOffset
      default 2;

    property NumGlyphs : TNumGlyphs
      read FNumGlyphs
      write SetNumGlyphs
      default 1;
  public
    constructor Create( AOwner : TComponent ); override;
    destructor Destroy; override;
  end; {== TRkCustomGlyphStatus Class Declaration ==}


  TRkGlyphStatus = class( TRkCustomGlyphStatus )
  published
    property Alignment;
    property Caption;
    property CaptionOffset;
    property Glyph;
    property GlyphAlignment;
    property GlyphOffset;
    property NumGlyphs;
  end;


  TRkPollingStatus = class( TRkCustomStatusControl )
  private
    FOnTimerExpired : TNotifyEvent;
    function GetActive : Boolean;
    procedure SetActive( Value : Boolean );
    function GetInterval : Word;
    procedure SetInterval( Value : Word );
  protected
    procedure TimerExpired; dynamic;
  public
    constructor Create( AOwner : TComponent ); override;
    destructor Destroy; override;
  published
    property Active : Boolean
      read GetActive
      write SetActive
      default True;

    property Interval : Word
      read GetInterval
      write SetInterval
      default 500;

    property OnTimerExpired : TNotifyEvent
      read FOnTimerExpired
      write FOnTimerExpired;
  end; {== TRkPollingStatus Class Declaration ==}


  TRkClockStatus = class( TRkPollingStatus )
  private
    FFormat : string;
    procedure SetFormat( Value : string );
  protected
    procedure TimerExpired; override;
    procedure UpdateCaption;
  public
    constructor Create( AOwner : TComponent ); override;
  published
    property Format : string
      read FFormat
      write SetFormat;

    { Inherited Properties }
    property Alignment;
    property CaptionOffset;
    property Width default 150;
  end; {== TRkClockStatus Class Declaration ==}


  TToggleKey      = ( tkCapsLock, tkNumLock, tkScrollLock );
  TToggleState    = ( tsOn, tsOff );

  TRkKeyStatus = class( TRkPollingStatus )
  private
    FKey : TToggleKey;
    FState : TToggleState;

    procedure SetKey( Value : TToggleKey );
    procedure SetState( Value : TToggleState );
  protected
    procedure TimerExpired; override;
    procedure Paint; override;
  public
    constructor Create( AOwner : TComponent ); override;

    property State : TToggleState
      read FState;
  published
    property Key : TToggleKey
      read FKey
      write SetKey
      default tkCapsLock;

    { Inherited Properties }
    property Width default 45;
  end; {== TRkKeyStatus Class Declaration ==}



implementation

uses
  RkPrgres;


type
  {=======================================================================
    TRkPollingTimer Class

    Implementation specific class designed to wrap a timer component and
    maintain a list of polling controls.  When the internal timer fires,
    the TimerExpired method of each polling control in the list is called.
  =======================================================================}

  TRkPollingTimer = class
  public
    Timer  : TTimer;
    Active : Boolean;
    PollingControls : TList;
    constructor Create;
    destructor Destroy; override;
    procedure TimerFired( Sender : TObject );
  end;

const
  VKCodes : array[ TToggleKey ] of Byte = ( vk_Capital,
                                            vk_NumLock,
                                            vk_Scroll );
  KeyStrings : array[ TToggleKey ] of string = ( 'CAPS', 'NUM', 'SCR' );

  PollingTimer : TRkPollingTimer = nil;



{====================================}
{== TRkCustomStatusControl Methods ==}
{====================================}

constructor TRkCustomStatusControl.Create( AOwner : TComponent );
begin
  inherited Create( AOwner );
  ControlStyle := ControlStyle + [ csOpaque ];
  FFrameStyle := fsRecessed;
  FCaptionOffset := 2;
  FBorderWidth := 2;

  Width := 100;
  Height := 20;
  ParentColor := True;
end;


procedure TRkCustomStatusControl.SetAlignment( Value : TAlignment );
begin
  if FAlignment <> Value then
  begin
    FAlignment := Value;
    Invalidate;
  end;
end;


procedure TRkCustomStatusControl.SetBorderWidth( Value : TBorderWidth );
begin
  if FBorderWidth <> Value then
  begin
    FBorderWidth := Value;
    Invalidate;                  // Invalidate b/c Border position changes
  end;
end;


procedure TRkCustomStatusControl.SetCaption( Value : string );
begin
  if FCaption <> Value then
  begin
    FCaption := Value;
    Invalidate;
  end;
end;


procedure TRkCustomStatusControl.SetCaptionOffset( Value : Integer );
begin
  if FCaptionOffset <> Value then
  begin
    FCaptionOffset := Value;
    Invalidate;
  end;
end;


procedure TRkCustomStatusControl.SetFrameStyle( Value : TFrameStyle );
begin
  if FFrameStyle <> Value then
  begin
    FFrameStyle := Value;
    Invalidate;
  end;
end;


procedure TRkCustomStatusControl.CMTextChanged( var Msg : TMessage );
begin
  Invalidate;
end;


function TRkCustomStatusControl.GetCaptionRect : TRect;
var
  Offset : Integer;
begin
  Result := ClientRect;

  if FFrameStyle in [ fsFlat, fsRecessed, fsRaised ] then
    Offset := FBorderWidth + 1
  else if FFrameStyle in [ fsGroove..fsBump ] then
    Offset := FBorderWidth + 2
  else
    Offset := FBorderWidth;

  InflateRect( Result, -Offset, -Offset );
end;


procedure TRkCustomStatusControl.DrawCaption( CaptionStr : string );
const
  SetTextAlignments : array[ TAlignment ] of Word = ( ta_Left,
                                                      ta_Right,
                                                      ta_Center );
var
  X, Y : Integer;
  R, TempRct : TRect;
begin
  if CaptionStr <> '' then
  begin
    Canvas.Font := Font;
    with Canvas do
    begin
      R := GetCaptionRect;

      // Set brush color so that old text in caption area gets erased
      Brush.Color := Color;

      X := 0;
      SetTextAlign( Canvas.Handle, SetTextAlignments[ FAlignment ] );
      case FAlignment of
        taLeftJustify:  X := R.Left + FCaptionOffset + 1;
        taRightJustify: X := R.Right - FCaptionOffset - 1;
        taCenter:       X := R.Left + ( R.Right - R.Left ) div 2;
      end;
      Y := R.Top + ( R.Bottom - R.Top - Canvas.TextHeight( 'Pp' ) ) div 2;

      if not Enabled then
      begin
        TempRct := R;
        OffsetRect( TempRct, 1, 1 );
        Font.Color := clBtnHighlight;
        TextRect( R, X + 1, Y + 1, CaptionStr );
        Font.Color := clBtnShadow;
        Brush.Style := bsClear;
      end;

      TextRect( R, X, Y, CaptionStr );
    end; { with }
  end;
end; {= TRkCustomStatusControl.DrawCaption =}


procedure TRkCustomStatusControl.DrawStatusBorder;
var
  R : TRect;
begin
  with Canvas do
  begin
    Brush.Color := Color;
    R := Rect( 0, 0, Width, Height );
    FillRect( R );
    Pen.Width := 1;
    InflateRect( R, -FBorderWidth, -FBorderWidth );
    DrawBorder( Canvas, R, FFrameStyle );
  end;
end; {= TRkCustomStatusControl.DrawStatusBorder =}


procedure TRkCustomStatusControl.Paint;
begin
  DrawStatusBorder;
  DrawCaption( Caption );
end;



{==================================}
{== TRkCustomGlyphStatus Methods ==}
{==================================}

constructor TRkCustomGlyphStatus.Create( AOwner : TComponent );
begin
  inherited Create( AOwner );
  FGlyph := TBitmap.Create;
  FGlyph.OnChange := GlyphChanged;
  FGlyphOffset := 2;
  FNumGlyphs := 1;
  FGlyphAlignment := gaLeft;
end;


destructor TRkCustomGlyphStatus.Destroy;
begin
  FGlyph.Free;                            // Must destroy the glyph object
  inherited Destroy;
end;


procedure TRkCustomGlyphStatus.SetGlyphAlignment(Value : TGlyphAlignment);
begin
  if FGlyphAlignment <> Value then
  begin
    FGlyphAlignment := Value;
    Invalidate;
  end;
end;


procedure TRkCustomGlyphStatus.SetGlyphOffset( Value : Integer );
begin
  if FGlyphOffset <> Value then
  begin
    FGlyphOffset := Value;
    Invalidate;
  end;
end;


procedure TRkCustomGlyphStatus.SetNumGlyphs( Value : TNumGlyphs );
begin
  if FNumGlyphs <> Value then
  begin
    FNumGlyphs := Value;
    Invalidate;
  end;
end;


procedure TRkCustomGlyphStatus.GlyphChanged( Sender : TObject );
var
  N : Integer;
begin
  if ( FGlyph.Height <> 0 ) and ( FGlyph.Width mod FGlyph.Height = 0) then
  begin
    // User has changed Glyph bitmap, recalculate the number of glyphs
    N := FGlyph.Width div FGlyph.Height;
    if N > 4 then
      N := 1;
    SetNumGlyphs( N );
  end;
  Invalidate;
end;


procedure TRkCustomGlyphStatus.SetGlyph( Value : TBitmap );
begin
  FGlyph.Assign( Value );
end;


function TRkCustomGlyphStatus.GetCaptionRect : TRect;
begin
  Result := inherited GetCaptionRect;

  if FGlyphAlignment = gaLeft then
    Inc( Result.Left, FGlyphOffset + FGlyph.Width div FNumGlyphs )
  else
    Dec( Result.Right, FGlyphOffset  + FGlyph.Width div FNumGlyphs );
end;


procedure TRkCustomGlyphStatus.Paint;
var
  DestRct, DrawRct, SrcRct, CaptionRct, GlyphRct : TRect;
  DestBmp : TBitmap;
  W, H, X, TopOffset : Integer;
  TransparentColor : TColor;
begin
  inherited Paint;

  Canvas.Font := Font;
  with Canvas do
  begin
    CaptionRct := GetCaptionRect;

    W := FGlyph.Width div FNumGlyphs;
    if ( CaptionRct.Bottom - CaptionRct.Top ) < FGlyph.Height then
      H := CaptionRct.Bottom - CaptionRct.Top
    else
      H := FGlyph.Height;

    DestRct := Rect( 0, 0, W, H );
    if ( FNumGlyphs > 1 ) and not Enabled then
      SrcRct := Rect( W, 0, W + W, H )
    else
      SrcRct := Rect( 0, 0, W, H );

    if FGlyphAlignment = gaLeft then
      X := FGlyphOffset + BorderWidth
    else
      X := Width - ( W + FGlyphOffset + BorderWidth + 2 );

    if FFrameStyle in [ fsGroove..fsBump ] then
      Inc( X, 2 )
    else if FFrameStyle in [ fsFlat, fsRecessed, fsRaised ] then
      Inc( X );

    DrawRct := Rect( X, 0, X + W, H );
    TopOffset := Abs( ( Height - H - BorderWidth - 1) ) div 2;
    OffsetRect( DrawRct, 0, BorderWidth + TopOffset );

    Brush.Color := Color;
    GlyphRct := inherited GetCaptionRect;
    GlyphRct.Right := CaptionRct.Left;
    FillRect( GlyphRct );

    { The DestBmp holds the desired region of the FGlyph bitmap }

    DestBmp := TBitmap.Create;
    try
      // Don't forget to set the Width and Height of destination bitmap
      DestBmp.Width := W;
      DestBmp.Height := H;
      DestBmp.Canvas.Brush.Color := Color;

      // Lower left pixel determines transparent color
      TransparentColor := FGlyph.Canvas.Pixels[ 0, H - 1 ];

      DestBmp.Canvas.BrushCopy(DestRct, FGlyph, SrcRct, TransparentColor);
      Draw( DrawRct.Left, DrawRct.Top, DestBmp );
    finally
      DestBmp.Free;
    end;
  end; { with }
end; {= TRkCustomGlyphStatus.Paint =}



{=============================}
{== TRkPollingTimer Methods ==}
{=============================}

constructor TRkPollingTimer.Create;
begin
  inherited Create;
  Timer := TTimer.Create( nil );
  Timer.Enabled := True;
  Timer.OnTimer := TimerFired;
  Timer.Interval := 500;

  PollingControls := TList.Create;
end;


destructor TRkPollingTimer.Destroy;
begin
  Timer.Free;
  PollingControls.Free;
  inherited Destroy;
end;


procedure TRkPollingTimer.TimerFired( Sender : TObject );
var
  I : Integer;
begin
  for I := 0 to PollingControls.Count - 1 do
    TRkPollingStatus( PollingControls.Items[ I ] ).TimerExpired;
end;



{==============================}
{== TRkPollingStatus Methods ==}
{==============================}

constructor TRkPollingStatus.Create( AOwner : TComponent );
begin
  inherited Create( AOwner );

  // If the PollingTimer object has not already been created,
  // then create it

  if PollingTimer = nil then
    PollingTimer := TRkPollingTimer.Create;
  PollingTimer.PollingControls.Add( Self );
end;


destructor TRkPollingStatus.Destroy;
begin
  // Delete current Polling Control from PollingControls list
  with PollingTimer.PollingControls do
    Delete( IndexOf( Self ) );

  if PollingTimer.PollingControls.Count = 0 then
  begin
    // If no more polling controls are left, destroy PollingTimer
    PollingTimer.Free;
    PollingTimer := nil;
  end;
  inherited Destroy;
end;


function TRkPollingStatus.GetActive : Boolean;
begin
  Result := PollingTimer.Timer.Enabled;
end;


procedure TRkPollingStatus.SetActive( Value : Boolean );
begin
  if PollingTimer.Timer.Enabled <> Value then
    PollingTimer.Timer.Enabled := Value;
end;


function TRkPollingStatus.GetInterval : Word;
begin
  Result := PollingTimer.Timer.Interval;
end;


procedure TRkPollingStatus.SetInterval( Value : Word );
begin
  if Value <> PollingTimer.Timer.Interval then
    PollingTimer.Timer.Interval := Value;
end;


procedure TRkPollingStatus.TimerExpired;
begin
  if Assigned( FOnTimerExpired ) then
    FOnTimerExpired( Self );
end;

{============================}
{== TRkClockStatus Methods ==}
{============================}

constructor TRkClockStatus.Create( AOwner : TComponent );
begin
  inherited Create( AOwner );
  Width := 150;
  FFormat := '';
  FCaption := '';
end;


procedure TRkClockStatus.UpdateCaption;
var
  S : string;
begin
  if FFormat <> '' then
    S := FormatDateTime( FFormat, Now )
  else
    S := DateTimeToStr( Now );
  if FCaption <> S then
  begin
    FCaption := S;
    Invalidate;
  end;
end;


procedure TRkClockStatus.TimerExpired;
begin
  UpdateCaption;
  inherited TimerExpired;
end;


procedure TRkClockStatus.SetFormat( Value : string );
begin
  if FFormat <> Value then
  begin
    FFormat := Value;
    UpdateCaption;
    Invalidate;
  end;
end;



{==========================}
{== TRkKeyStatus Methods ==}
{==========================}

constructor TRkKeyStatus.Create( AOwner : TComponent );
begin
  inherited Create( AOwner );
  Width := 45;
  Alignment := taCenter;
end;


procedure TRkKeyStatus.SetKey( Value : TToggleKey );
begin
  if FKey <> Value then
  begin
    FKey := Value;
    Invalidate;
  end;
end;


procedure TRkKeyStatus.SetState( Value : TToggleState );
begin
  if FState <> Value then
  begin
    FState := Value;
    Invalidate;
  end;
end;


procedure TRkKeyStatus.TimerExpired;
begin
  // If the low-order bit is 1, the key is active (or On)
  if ( GetKeyState( VKCodes[ FKey ] ) and $1 ) = $1 then
    SetState( tsOn )
  else
    SetState( tsOff );
  inherited TimerExpired;
end;


procedure TRkKeyStatus.Paint;
begin
  Enabled := FState = tsOn;
  Caption := KeyStrings[ FKey ];
  inherited Paint;
end;


end.

