{=========================================================================
  RkTabEdt Unit

  This unit implements a property editor for the TabStop property
  associated with the TRkTabbedListBox component.  This unit defines the
  TRkTabStopProperty class and the TRkTabStopEditDlg class.  The property
  editor class displays the dialog in its Edit method.  The dialog allows
  the user to modify the TabStop property at design-time.

  Developing Custom Delphi 3 Components
  Copyright  1995-1997 by Ray Konopka
=========================================================================}

unit RkTabEdt;

interface

uses
  SysUtils, Windows, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, Buttons, RkTabLst, DsgnIntf, RkTrkBar;

type
  TRkTabStopProperty = class( TPropertyEditor )
    function GetAttributes : TPropertyAttributes; override;
    function GetValue : string; override;
    procedure Edit; override;
  end;


  TRkTabStopEditDlg = class( TForm )
    BtnOK : TButton;
    BtnCancel : TButton;
    GrpPreview : TGroupBox;
    GrpTabStops : TGroupBox;
    BtnAdd : TBitBtn;
    LstTabs : TListBox;
    BtnDelete : TBitBtn;
    LblMin : TLabel;
    LblMax : TLabel;
    Label3 : TLabel;
    LblTabNum : TLabel;
    LstPreview : TRkTabbedListBox;
    TrkTabPos : TRkTrackBar;
    procedure BtnAddClick( Sender : TObject );
    procedure BtnDeleteClick( Sender : TObject );
    procedure LstTabsClick( Sender : TObject );
    procedure FormShow( Sender : TObject );
    procedure TrkTabPosChange( Sender : TObject );
    procedure FormCreate( Sender : TObject );
  private
    FUpdating : Boolean;
  end;


procedure Register;

implementation

{$R *.DFM}


{========================}
{== TRkTabStopProperty ==}
{========================}

function TRkTabStopProperty.GetAttributes : TPropertyAttributes;
begin
  Result := [ paDialog ];        // This is a dialog-based property editor
end;


function TRkTabStopProperty.GetValue : string;
begin
  // Display the property's type name in the Object Inspector
  Result := Format( '(%s)', [ GetPropType^.Name ] );
end;


{=========================================================================
  TRkTabStopProperty.Edit

  This method will be invoked by Delphi when the user double-clicks on
  the TabStop property line or clicks the ellipsis button. This method
  creates an instance of the TRkTabStopEditDlg form and initializes the
  preview list box according to the tabbed list box being edited. Next,
  the edit form is displayed, and when the user closes the form, the new
  tab stops, if any, are recorded in the new property value.
=========================================================================}

procedure TRkTabStopProperty.Edit;
var
  D : TRkTabStopEditDlg;
  I : Integer;
begin
  D := TRkTabStopEditDlg.Create( Application );
  try
    // Under Delphi 3.0, GetComponent returns a TPersistent reference
    // while under Delphi 2.0 it returns a TComponent reference.
     
    D.Caption := TComponent( GetComponent(0) ).Owner.Name + '.' +
                 TComponent( GetComponent(0) ).Name + '.' +
                 GetName + ' - ' + D.Caption;

    // Set the preview list box strings and font to be same as component's
    D.LstPreview.Items := TRkTabbedListBox( GetComponent( 0 ) ).Items;
    D.LstPreview.Font := TRkTabbedListBox( GetComponent( 0 ) ).Font;

    // Set the preview list box tab stops to be same as current prop value
    D.LstPreview.TabStops := TRkTabStopList( GetOrdValue );

    // Add any existing tab stops to the LstTabs list box
    for I := 0 to D.LstPreview.TabStops.Count - 1 do
      D.LstTabs.Items.Add( IntToStr( D.LstPreview.TabStops[ I ] ) );

    if D.ShowModal = mrOK then                  // Display property dialog
    begin
      // If user presses OK, update the property with the tab stops
      // stored in the preview list.

      SetOrdValue( Longint( D.LstPreview.TabStops ) );
    end;
  finally
    D.Free;
  end;
end; {= TRkTabStopProperty.Edit =}



{===============================}
{== TRkTabStopEditDlg Methods ==}
{===============================}

procedure TRkTabStopEditDlg.FormCreate( Sender : TObject );
begin
  FUpdating := False;
end;


procedure TRkTabStopEditDlg.BtnAddClick( Sender : TObject );
var
  NewTab : Word;
  Index : Integer;
begin
  NewTab := 8;
  if LstPreview.TabStops.Count > 0 then
  begin
    // Add a new tab stop 8 positions after the last tab stop
    NewTab := LstPreview.TabStops[ LstPreview.TabStops.Count - 1 ] + 8;
  end;
  // Add the new tab stop to the LstTabs list
  Index := LstTabs.Items.Add( IntToStr( NewTab ) );

  // Add the new tab stop to the preview list
  LstPreview.TabStops.Add( NewTab );

  LstTabs.ItemIndex := Index;           // Select the newly added tab stop
  LstTabsClick( nil );                             // Update the track bar
  BtnDelete.Enabled := True;
  TrkTabPos.Enabled := True;
end;


procedure TRkTabStopEditDlg.BtnDeleteClick( Sender : TObject );
var
  Index : Integer;
begin
  Index := LstTabs.ItemIndex;

  // Remove selected tab stop from preview list
  LstPreview.TabStops.Delete( Index );

  LstTabs.Items.Delete( Index );

  if LstTabs.Items.Count > 0 then
  begin
    if Index = LstTabs.Items.Count then
      Dec( Index );
    LstTabs.ItemIndex := Index;
    LstTabsClick( nil );                           // Update the track bar
  end
  else
  begin
    BtnDelete.Enabled := False;
    TrkTabPos.Enabled := False;
  end;
end;


{=========================================================================
  TRkTabStopEditDlg.LstTabsClick

  This method updates the Min and Max values of the track bar to reflect
  the range a particular tab stop may move.  A tab stop may not move past
  a tab stop that exists before or after it in the list.  Setting the
  track bar range ensures that this cannot happen. The FUpdating flag is
  set at the beginning of this method to prevent a change in the
  TrackBar's Min or Max value from causing the TrkTabPosChange event from
  altering the display.
=========================================================================}

procedure TRkTabStopEditDlg.LstTabsClick( Sender : TObject );
begin
  FUpdating := True;
  TrkTabPos.Min := 0;
  TrkTabPos.Max := 100;

  if LstTabs.ItemIndex > 0 then
    // Get previous tab stop value
    TrkTabPos.Min := StrToInt( LstTabs.Items[ LstTabs.ItemIndex - 1 ] )
  else
    TrkTabPos.Min := 0;

  if LstTabs.ItemIndex < LstTabs.Items.Count - 1 then
    // Get next tab stop value
    TrkTabPos.Max := StrToInt( LstTabs.Items[ LstTabs.ItemIndex + 1 ] )
  else
    TrkTabPos.Max := 100;

  // Update TrackBar position to reflect currently selected tab stop }
  TrkTabPos.Position := LstPreview.TabStops[ LstTabs.ItemIndex ];
  LblTabNum.Caption := IntToStr( LstTabs.ItemIndex + 1 );
  LblMin.Caption := IntToStr( TrkTabPos.Min );
  LblMax.Caption := IntToStr( TrkTabPos.Max );
  FUpdating := False;
end;


procedure TRkTabStopEditDlg.FormShow( Sender : TObject );
begin
  if LstTabs.Items.Count > 0 then
  begin
    LstTabs.ItemIndex := 0;
    LstTabsClick( nil );
    BtnDelete.Enabled := True;
    TrkTabPos.Enabled := True;
  end;
end;



{=========================================================================
  TRkTabStopEditDlg.TrkTabPosChange

  As the track bar is moved, the value of the selected tab stop is
  updated. The change is immediately reflected in the tab and preview
  list boxes.
=========================================================================}

procedure TRkTabStopEditDlg.TrkTabPosChange( Sender : TObject );
var
  Index : Integer;
begin
  if not FUpdating then
  begin
    Index := LstTabs.ItemIndex;
    LstPreview.TabStops[ Index ] := TrkTabPos.Position;
    LstTabs.Items[ Index ] := IntToStr( TrkTabPos.Position );
    LstTabs.ItemIndex := Index;
  end;
end;


{========================}
{== Register Procedure ==}
{========================}

procedure Register;
begin
  // This property editor is registered for any property of type
  // TRkTabStopList. This includes the TabStops property for the
  // TRkTabbedListBox component.

  RegisterPropertyEditor( TypeInfo( TRkTabStopList ), nil,
                          '', TRkTabStopProperty );
end;

end.
